import { DateComponent, memoize, mapHash, NowTimer } from '@fullcalendar/core/internal.js';
import { createElement } from '@fullcalendar/core/preact.js';
import { createDayHeaderFormatter, DayTableSlicer } from '@fullcalendar/daygrid/internal.js';
import { buildResourceRowConfigs, ResourceDayTableJoiner } from '@fullcalendar/resource-daygrid/internal.js';
import { VResourceJoiner, flattenResources, VResourceSplitter, DEFAULT_RESOURCE_ORDER, ResourcelessDayTableModel, DayResourceTableModel, ResourceDayTableModel } from '@fullcalendar/resource/internal.js';
import { AllDaySplitter, buildDayRanges, organizeSegsByCol, splitInteractionByCol, DayTimeColsSlicer, TimeGridLayout, buildTimeColsModel } from '@fullcalendar/timegrid/internal.js';

class ResourceDayTimeColsJoiner extends VResourceJoiner {
    transformSeg(seg, resourceDayTable, resourceI) {
        return [
            Object.assign(Object.assign({}, seg), { col: resourceDayTable.computeCol(seg.col, resourceI) }),
        ];
    }
}

class ResourceTimeGridView extends DateComponent {
    constructor() {
        super(...arguments);
        // memo
        this.flattenResources = memoize(flattenResources);
        this.buildResourceTimeColsModel = memoize(buildResourceTimeColsModel);
        this.buildResourceRowConfigs = memoize(buildResourceRowConfigs);
        this.createDayHeaderFormatter = memoize(createDayHeaderFormatter);
        // internal
        this.allDaySplitter = new AllDaySplitter();
        // for all-day-resource props
        this.allDayResourceSplitter = new VResourceSplitter();
        this.allDayResourceSlicers = {};
        this.allDayResourceJoiner = new ResourceDayTableJoiner();
        // for timed resource props
        this.buildDayRanges = memoize(buildDayRanges);
        this.timedResourceSplitter = new VResourceSplitter();
        this.timedResourceSlicers = {};
        this.timedResourceJoiner = new ResourceDayTimeColsJoiner();
        // timed-only column splitting
        this.splitFgEventSegs = memoize(organizeSegsByCol);
        this.splitBgEventSegs = memoize(organizeSegsByCol);
        this.splitBusinessHourSegs = memoize(organizeSegsByCol);
        this.splitNowIndicatorSegs = memoize(organizeSegsByCol);
        this.splitDateSelectionSegs = memoize(organizeSegsByCol);
        this.splitEventDrag = memoize(splitInteractionByCol);
        this.splitEventResize = memoize(splitInteractionByCol);
        this.isHitComboAllowed = (hit0, hit1) => {
            let allowAcrossResources = this.dayRanges.length === 1;
            return allowAcrossResources || hit0.dateSpan.resourceId === hit1.dateSpan.resourceId;
        };
    }
    render() {
        let { props, context } = this;
        let { options, dateEnv } = context;
        let { dateProfile } = props;
        let resourceOrderSpecs = options.resourceOrder || DEFAULT_RESOURCE_ORDER;
        let resources = this.flattenResources(props.resourceStore, resourceOrderSpecs);
        let resourceDayTableModel = this.buildResourceTimeColsModel(dateProfile, context.dateProfileGenerator, resources, options.datesAboveResources, context);
        // split seg by all-day/timed
        let splitProps = this.allDaySplitter.splitProps(props);
        // split the all-day segs by resource
        let allDayResourceSplitProps = this.allDayResourceSplitter.splitProps({
            businessHours: splitProps.allDay.businessHours,
            dateSelection: splitProps.allDay.dateSelection,
            eventStore: splitProps.allDay.eventStore,
            eventUiBases: splitProps.allDay.eventUiBases,
            eventSelection: splitProps.allDay.eventSelection,
            eventDrag: splitProps.allDay.eventDrag,
            eventResize: splitProps.allDay.eventResize,
            resourceDayTableModel,
        });
        this.allDayResourceSlicers = mapHash(allDayResourceSplitProps, (split, resourceId) => this.allDayResourceSlicers[resourceId] || new DayTableSlicer());
        let allDayResourceSlicedProps = mapHash(this.allDayResourceSlicers, (slicer, resourceId) => slicer.sliceProps(allDayResourceSplitProps[resourceId], dateProfile, options.nextDayThreshold, context, resourceDayTableModel.dayTableModel));
        let allDayResourceJoinedProps = this.allDayResourceJoiner.joinProps(allDayResourceSlicedProps, resourceDayTableModel);
        // split the timed segs by resource
        let dayRanges = this.dayRanges = this.buildDayRanges(resourceDayTableModel.dayTableModel, dateProfile, dateEnv);
        let timedResourceSplitProps = this.timedResourceSplitter.splitProps({
            businessHours: splitProps.timed.businessHours,
            dateSelection: splitProps.timed.dateSelection,
            eventStore: splitProps.timed.eventStore,
            eventUiBases: splitProps.timed.eventUiBases,
            eventSelection: splitProps.timed.eventSelection,
            eventDrag: splitProps.timed.eventDrag,
            eventResize: splitProps.timed.eventResize,
            resourceDayTableModel,
        });
        this.timedResourceSlicers = mapHash(timedResourceSplitProps, (split, resourceId) => this.timedResourceSlicers[resourceId] || new DayTimeColsSlicer());
        let timedResourceSlicedProps = mapHash(this.timedResourceSlicers, (slicer, resourceId) => slicer.sliceProps(timedResourceSplitProps[resourceId], dateProfile, null, context, dayRanges));
        let timedResourceJoinedProps = this.timedResourceJoiner.joinProps(timedResourceSlicedProps, resourceDayTableModel);
        let datesRepDistinctDays = resourceDayTableModel.dayTableModel.rowCnt === 1;
        let dayHeaderFormat = this.createDayHeaderFormatter(context.options.dayHeaderFormat, datesRepDistinctDays, resourceDayTableModel.colCnt);
        return (createElement(NowTimer, { unit: options.nowIndicator ? 'minute' : 'day' /* hacky */ }, (nowDate, todayRange) => {
            // timed-only column splitting
            let colCnt = resourceDayTableModel.colCnt;
            let fgEventSegsByCol = this.splitFgEventSegs(timedResourceJoinedProps.fgEventSegs, colCnt);
            let bgEventSegsByCol = this.splitBgEventSegs(timedResourceJoinedProps.bgEventSegs, colCnt);
            let businessHourSegsByCol = this.splitBusinessHourSegs(timedResourceJoinedProps.businessHourSegs, colCnt);
            let nowIndicatorSegsByCol = this.splitNowIndicatorSegs((() => {
                // was buildNowIndicatorSegs
                let nonResourceSegs = options.nowIndicator
                    ? this.timedResourceSlicers[''].sliceNowDate(nowDate, this.props.dateProfile, this.context.options.nextDayThreshold, this.context, this.dayRanges)
                    : []; // TODO: breaks memoization?
                return this.timedResourceJoiner.expandSegs(resourceDayTableModel, nonResourceSegs);
            })(), colCnt);
            let dateSelectionSegsByCol = this.splitDateSelectionSegs(timedResourceJoinedProps.dateSelectionSegs, colCnt);
            let eventDragByCol = this.splitEventDrag(timedResourceJoinedProps.eventDrag, colCnt);
            let eventResizeByCol = this.splitEventResize(timedResourceJoinedProps.eventResize, colCnt);
            const headerTiers = this.buildResourceRowConfigs(resources, options.datesAboveResources, resourceDayTableModel.dayTableModel.headerDates, datesRepDistinctDays, props.dateProfile, todayRange, dayHeaderFormat, context);
            return (createElement(TimeGridLayout, { labelId: props.labelId, labelStr: props.labelStr, dateProfile: dateProfile, nowDate: nowDate, todayRange: todayRange, cells: resourceDayTableModel.cells[0], forPrint: props.forPrint, isHitComboAllowed: this.isHitComboAllowed, className: 'fc-resource-timegrid fc-timegrid', 
                // header content
                headerTiers: headerTiers, 
                // all-day content
                fgEventSegs: allDayResourceJoinedProps.fgEventSegs, bgEventSegs: allDayResourceJoinedProps.bgEventSegs, businessHourSegs: allDayResourceJoinedProps.businessHourSegs, dateSelectionSegs: allDayResourceJoinedProps.dateSelectionSegs, eventDrag: allDayResourceJoinedProps.eventDrag, eventResize: allDayResourceJoinedProps.eventResize, 
                // timed content
                fgEventSegsByCol: fgEventSegsByCol, bgEventSegsByCol: bgEventSegsByCol, businessHourSegsByCol: businessHourSegsByCol, nowIndicatorSegsByCol: nowIndicatorSegsByCol, dateSelectionSegsByCol: dateSelectionSegsByCol, eventDragByCol: eventDragByCol, eventResizeByCol: eventResizeByCol, 
                // universal content
                eventSelection: allDayResourceJoinedProps.eventSelection || splitProps.timed.eventSelection }));
        }));
    }
}
/*
TODO: kill this and DayResourceTableModel/ResourceDayTableModel
*/
function buildResourceTimeColsModel(dateProfile, dateProfileGenerator, resources, datesAboveResources, context) {
    let dayTable = buildTimeColsModel(dateProfile, dateProfileGenerator);
    if (!resources.length) {
        return new ResourcelessDayTableModel(dayTable);
    }
    return datesAboveResources ?
        new DayResourceTableModel(dayTable, resources, context) :
        new ResourceDayTableModel(dayTable, resources, context);
}

export { ResourceTimeGridView };
